@@ -2,22 +2,23 @@
 #this repository contains the full copyright notices and license terms.
 from decimal import Decimal
 import datetime
-import time
 from itertools import groupby
 from operator import itemgetter
+from sql.aggregate import Sum
+from sql.conditionals import Coalesce
 
 from trytond.model import ModelView, ModelSQL, fields
 from trytond.wizard import Wizard, StateTransition, StateView, StateAction, \
     Button
 from trytond.report import Report
-from trytond.backend import TableHandler, FIELDS
+from trytond import backend
 from trytond.pyson import Eval, PYSONEncoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 from trytond.rpc import RPC
 from trytond.tools import reduce_ids
 
-__all__ = ['Move', 'Reconciliation', 'Line', 'Move2', 'OpenJournalAsk',
+__all__ = ['Move', 'Reconciliation', 'Line', 'OpenJournalAsk',
     'OpenJournal', 'OpenAccount', 'ReconcileLinesWriteOff', 'ReconcileLines',
     'UnreconcileLinesStart', 'UnreconcileLines', 'OpenReconcileLinesStart',
     'OpenReconcileLines', 'FiscalYearLine', 'FiscalYear2',
@@ -79,10 +80,6 @@ class Move(ModelSQL, ModelView):
                     '"%(move)s" to draft in journal "%(journal)s".'),
                 'modify_posted_move': ('You can not modify move "%s" because '
                     'it is already posted.'),
-                'period_centralized_journal': ('Move "%(move)s" cannot be '
-                    'created because there is already a move in journal '
-                    '"%(journal)s" and you cannot create more than one move '
-                    'per period in a centralized journal.'),
                 'company_in_move': ('You can not create lines on accounts'
                     'of different companies in move "%s".'),
                 'date_outside_period': ('You can not create move "%(move)s" '
@@ -101,6 +98,7 @@ class Move(ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
         table = TableHandler(cursor, cls, module_name)
 
@@ -157,7 +155,7 @@ class Move(ModelSQL, ModelView):
     @classmethod
     def _get_origin(cls):
         'Return list of Model names for origin Reference'
-        return []
+        return ['account.fiscalyear']
 
     @classmethod
     def get_origin(cls):
@@ -172,23 +170,9 @@ class Move(ModelSQL, ModelView):
     def validate(cls, moves):
         super(Move, cls).validate(moves)
         for move in moves:
-            move.check_centralisation()
             move.check_company()
             move.check_date()
 
-    def check_centralisation(self):
-        if self.journal.centralised:
-            moves = self.search([
-                    ('period', '=', self.period.id),
-                    ('journal', '=', self.journal.id),
-                    ('state', '!=', 'posted'),
-                    ], limit=2)
-            if len(moves) > 1:
-                self.raise_user_error('period_centralized_journal', {
-                        'move': self.rec_name,
-                        'journal': self.journal.rec_name,
-                        })
-
     def check_company(self):
         company_id = -1
         for line in self.lines:
@@ -200,9 +184,7 @@ class Move(ModelSQL, ModelView):
     def check_date(self):
         if (self.date < self.period.start_date
                 or self.date > self.period.end_date):
-            self.raise_user_error('date_outside_period', {
-                        'move': self.rec_name,
-                        })
+            self.raise_user_error('date_outside_period', (self.rec_name,))
 
     @classmethod
     def check_modify(cls, moves):
@@ -233,7 +215,6 @@ class Move(ModelSQL, ModelView):
     @classmethod
     def create(cls, vlist):
         pool = Pool()
-        MoveLine = pool.get('account.move.line')
         Sequence = pool.get('ir.sequence')
         Journal = pool.get('account.journal')
 
@@ -247,15 +228,6 @@ class Move(ModelSQL, ModelView):
                     vals['number'] = Sequence.get_id(journal.sequence.id)
 
         moves = super(Move, cls).create(vlist)
-        for move in moves:
-            if move.journal.centralised:
-                line, = MoveLine.create([{
-                            'account': move.journal.credit_account.id,
-                            'move': move.id,
-                            }])
-                cls.write([move], {
-                        'centralised_line': line.id,
-                        })
         cls.validate_move(moves)
         return moves
 
@@ -291,11 +263,12 @@ class Move(ModelSQL, ModelView):
     @classmethod
     def validate_move(cls, moves):
         '''
-        Validate balanced move and centralise it if in centralised journal
+        Validate balanced move
         '''
         pool = Pool()
         MoveLine = pool.get('account.move.line')
         User = pool.get('res.user')
+        line = MoveLine.__table__()
 
         cursor = Transaction().cursor
 
@@ -310,18 +283,17 @@ class Move(ModelSQL, ModelView):
         for i in range(0, len(moves), cursor.IN_MAX):
             sub_moves = moves[i:i + cursor.IN_MAX]
             sub_move_ids = [m.id for m in sub_moves]
-            red_sql, red_ids = reduce_ids('move', sub_move_ids)
+            red_sql = reduce_ids(line.move, sub_move_ids)
 
-            cursor.execute('SELECT move, SUM(debit - credit) '
-                'FROM "' + MoveLine._table + '" '
-                'WHERE ' + red_sql + ' '
-                'GROUP BY move', red_ids)
+            cursor.execute(*line.select(line.move,
+                    Sum(line.debit - line.credit),
+                    where=red_sql,
+                    group_by=line.move))
             amounts.update(dict(cursor.fetchall()))
 
-            cursor.execute('SELECT move, id '
-                'FROM "' + MoveLine._table + '" '
-                'WHERE ' + red_sql + ' AND state = %s '
-                'ORDER BY move', red_ids + ['draft'])
+            cursor.execute(*line.select(line.move, line.id,
+                    where=red_sql & (line.state == 'draft'),
+                    order_by=line.move))
             move2draft_lines.update(dict((k, (j[1] for j in g))
                     for k, g in groupby(cursor.fetchall(), itemgetter(0))))
 
@@ -337,39 +309,7 @@ class Move(ModelSQL, ModelView):
             draft_lines = MoveLine.browse(
                 list(move2draft_lines.get(move.id, [])))
             if not company.currency.is_zero(amount):
-                if not move.journal.centralised:
-                    draft_moves.append(move.id)
-                else:
-                    if not move.centralised_line:
-                        centralised_amount = - amount
-                    else:
-                        centralised_amount = move.centralised_line.debit \
-                            - move.centralised_line.credit \
-                            - amount
-                    if centralised_amount >= Decimal('0.0'):
-                        debit = centralised_amount
-                        credit = Decimal('0.0')
-                        account_id = move.journal.debit_account.id
-                    else:
-                        debit = Decimal('0.0')
-                        credit = - centralised_amount
-                        account_id = move.journal.credit_account.id
-                    if not move.centralised_line:
-                        centralised_line, = MoveLine.create([{
-                                    'debit': debit,
-                                    'credit': credit,
-                                    'account': account_id,
-                                    'move': move.id,
-                                    }])
-                        cls.write([move], {
-                            'centralised_line': centralised_line.id,
-                            })
-                    else:
-                        MoveLine.write([move.centralised_line], {
-                                'debit': debit,
-                                'credit': credit,
-                                'account': account_id,
-                                })
+                draft_moves.append(move.id)
                 continue
             if not draft_lines:
                 continue
@@ -381,11 +321,12 @@ class Move(ModelSQL, ModelView):
             if move_ids:
                 for i in range(0, len(move_ids), cursor.IN_MAX):
                     sub_ids = move_ids[i:i + cursor.IN_MAX]
-                    red_sql, red_ids = reduce_ids('move', sub_ids)
+                    red_sql = reduce_ids(line.move, sub_ids)
                     # Use SQL to prevent double validate loop
-                    cursor.execute('UPDATE "' + MoveLine._table + '" '
-                        'SET state = %s '
-                        'WHERE ' + red_sql, [state] + red_ids)
+                    cursor.execute(*line.update(
+                            columns=[line.state],
+                            values=[state],
+                            where=red_sql))
 
     @classmethod
     @ModelView.button
@@ -393,6 +334,7 @@ class Move(ModelSQL, ModelView):
         pool = Pool()
         Sequence = pool.get('ir.sequence')
         Date = pool.get('ir.date')
+        Line = pool.get('account.move.line')
 
         for move in moves:
             amount = Decimal('0.0')
@@ -415,6 +357,12 @@ class Move(ModelSQL, ModelView):
                     move.period.post_move_sequence_used.id)
             cls.write([move], values)
 
+            to_reconcile = [l for l in move.lines
+                if ((line.debit == line.credit == Decimal('0'))
+                    and line.account.reconcile)]
+            if to_reconcile:
+                Line.reconcile(to_reconcile)
+
     @classmethod
     @ModelView.button
     def draft(cls, moves):
@@ -451,7 +399,8 @@ class Reconciliation(ModelSQL, ModelView):
                 'reconciliation_different_accounts': ('You can not reconcile '
                     'line "%(line)s" because it\'s account "%(account1)s" is '
                     'different from "%(account2)s".'),
-                'reconciliation_account_no_reconcile': ('You can not reconcile '
+                'reconciliation_account_no_reconcile': (
+                    'You can not reconcile '
                     'line "%(line)s" because it\'s account "%(account)s" is '
                     'configured as not reconcilable.'),
                 'reconciliation_different_parties': ('You can not reconcile '
@@ -603,6 +552,7 @@ class Line(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(Line, cls).__setup__()
+        cls._check_modify_exclude = ['reconciliation']
         cls._sql_constraints += [
             ('credit_debit',
                 'CHECK(credit * debit = 0.0)',
@@ -618,19 +568,20 @@ class Line(ModelSQL, ModelView):
                     'add/modify lines in closed journal period "%s".'),
                 'modify_posted_move': ('You can not modify lines of move "%s" '
                     'because it is already posted.'),
-                'modify_reconciled': ('You can not modify line "%s" because it '
-                    'is reconciled.'),
+                'modify_reconciled': ('You can not modify line "%s" because '
+                    'it is reconciled.'),
                 'no_journal': ('Move line cannot be created because there is '
                     'no journal defined.'),
                 'move_view_account': ('You can not create a move line with '
                     'account "%s" because it is a view account.'),
-                'move_inactive_account': ('You can not create a move line with '
-                    'account "%s" because it is inactive.'),
+                'move_inactive_account': ('You can not create a move line '
+                    'with account "%s" because it is inactive.'),
                 'already_reconciled': 'Line "%s" (%d) already reconciled.',
                 })
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
         table = TableHandler(cursor, cls, module_name)
 
@@ -991,18 +942,15 @@ class Line(ModelSQL, ModelView):
             return res
 
         if self.party and (not self.debit) and (not self.credit):
-            type_name = FIELDS[self.__class__.debit._type].sql_type(
-                self.__class__.debit)[0]
-            query = ('SELECT '
-                'CAST(COALESCE(SUM('
-                        '(COALESCE(debit, 0) - COALESCE(credit, 0))'
-                    '), 0) AS ' + type_name + ') '
-                'FROM account_move_line '
-                'WHERE reconciliation IS NULL '
-                    'AND party = %s '
-                    'AND account = %s')
-            cursor.execute(query,
-                (self.party.id, self.party.account_receivable.id))
+            type_name = self.__class__.debit.sql_type().base
+            table = self.__table__()
+            column = Coalesce(Sum(Coalesce(table.debit, 0)
+                    - Coalesce(table.credit, 0)), 0).cast(type_name)
+            where = ((table.reconciliation == None)
+                & (table.party == self.party.id))
+            cursor.execute(*table.select(column,
+                    where=where
+                    & (table.account == self.party.account_receivable.id)))
             amount = cursor.fetchone()[0]
             # SQLite uses float for SUM
             if not isinstance(amount, Decimal):
@@ -1020,8 +968,9 @@ class Line(ModelSQL, ModelView):
                 res['account.rec_name'] = \
                     self.party.account_receivable.rec_name
             else:
-                cursor.execute(query,
-                    (self.party.id, self.party.account_payable.id))
+                cursor.execute(*table.select(column,
+                        where=where
+                        & (table.account == self.party.account_payable.id)))
                 amount = cursor.fetchone()[0]
                 if not self.party.account_payable.currency.is_zero(amount):
                     if amount > Decimal('0.0'):
@@ -1108,17 +1057,42 @@ class Line(ModelSQL, ModelView):
             name = 'state'
         return [('move.' + name,) + tuple(clause[1:])]
 
+    def _order_move_field(name):
+        def order_field(tables):
+            pool = Pool()
+            Move = pool.get('account.move')
+            field = Move._fields[name]
+            table, _ = tables[None]
+            move_tables = tables.get('move')
+            if move_tables is None:
+                move = Move.__table__()
+                move_tables = {
+                    None: (move, move.id == table.move),
+                    }
+                tables['move'] = move_tables
+            return field.convert_order(name, move_tables, Move)
+        return staticmethod(order_field)
+    order_journal = _order_move_field('journal')
+    order_period = _order_move_field('period')
+    order_date = _order_move_field('date')
+    order_origin = _order_move_field('origin')
+    order_move_state = _order_move_field('state')
+
     @classmethod
-    def query_get(cls, obj='l'):
+    def query_get(cls, table):
         '''
         Return SQL clause and fiscal years for account move line
         depending of the context.
-        obj is the SQL alias of account_move_line in the query
+        table is the SQL instance of account.move.line table
         '''
-        FiscalYear = Pool().get('account.fiscalyear')
+        pool = Pool()
+        FiscalYear = pool.get('account.fiscalyear')
+        Move = pool.get('account.move')
+        Period = pool.get('account.period')
+        move = Move.__table__()
+        period = Period.__table__()
 
         if Transaction().context.get('date'):
-            time.strptime(str(Transaction().context['date']), '%Y-%m-%d')
             fiscalyears = FiscalYear.search([
                     ('start_date', '<=', Transaction().context['date']),
                     ('end_date', '>=', Transaction().context['date']),
@@ -1127,83 +1101,71 @@ class Line(ModelSQL, ModelView):
             fiscalyear_id = fiscalyears and fiscalyears[0].id or 0
 
             if Transaction().context.get('posted'):
-                return (obj + '.state != \'draft\' '
-                        'AND ' + obj + '.move IN ('
-                            'SELECT m.id FROM account_move AS m, '
-                                'account_period AS p '
-                                'WHERE m.period = p.id '
-                                    'AND p.fiscalyear = ' +
-                                        str(fiscalyear_id) + ' '
-                                    'AND m.date <= date(\'' +
-                                        str(Transaction().context['date']) +
-                                        '\') '
-                                    'AND m.state = \'posted\' '
-                            ')', [f.id for f in fiscalyears])
+                return ((table.state != 'draft')
+                    & table.move.in_(move.join(period,
+                            condition=move.period == period.id
+                            ).select(move.id,
+                            where=(move.fiscalyear == fiscalyear_id)
+                            & (move.date <= Transaction().context['date'])
+                            & (move.state == 'posted'))),
+                    [f.id for f in fiscalyears])
             else:
-                return (obj + '.state != \'draft\' '
-                        'AND ' + obj + '.move IN ('
-                            'SELECT m.id FROM account_move AS m, '
-                                'account_period AS p '
-                                'WHERE m.period = p.id '
-                                    'AND p.fiscalyear = ' +
-                                        str(fiscalyear_id) + ' '
-                                    'AND m.date <= date(\'' +
-                                        str(Transaction().context['date']) +
-                                        '\')'
-                            ')', [f.id for f in fiscalyears])
+                return ((table.state != 'draft')
+                    & table.move.in_(move.join(period,
+                            condition=move.period == period.id
+                            ).select(move.id,
+                            where=(period.fiscalyear == fiscalyear_id)
+                            & (move.date <= Transaction().context['date']))),
+                    [f.id for f in fiscalyears])
 
         if Transaction().context.get('periods'):
             if Transaction().context.get('fiscalyear'):
-                fiscalyear_ids = [int(Transaction().context['fiscalyear'])]
+                fiscalyear_ids = [Transaction().context['fiscalyear']]
             else:
                 fiscalyear_ids = []
-            ids = ','.join(
-                    str(int(x)) for x in Transaction().context['periods'])
             if Transaction().context.get('posted'):
-                return (obj + '.state != \'draft\' '
-                    'AND ' + obj + '.move IN ('
-                        'SELECT id FROM account_move '
-                        'WHERE period IN (' + ids + ') '
-                            'AND state = \'posted\' '
-                        ')', fiscalyear_ids)
+                return ((table.state != 'draft')
+                    & table.move.in_(
+                            move.select(move.id,
+                                where=period.in_(
+                                    Transaction().context['periods'])
+                                & move.state == 'posted')),
+                    fiscalyear_ids)
             else:
-                return (obj + '.state != \'draft\' '
-                    'AND ' + obj + '.move IN ('
-                        'SELECT id FROM account_move '
-                        'WHERE period IN (' + ids + ')'
-                        ')', fiscalyear_ids)
+                return ((table.state != 'draft')
+                    & table.move.in_(
+                        move.select(move.id,
+                            where=move.period.in_(
+                                Transaction().context['periods']))),
+                    fiscalyear_ids)
         else:
             if not Transaction().context.get('fiscalyear'):
                 fiscalyears = FiscalYear.search([
                     ('state', '=', 'open'),
                     ])
-                fiscalyear_ids = [f.id for f in fiscalyears]
-                fiscalyear_clause = (','.join(
-                        str(f.id) for f in fiscalyears)) or '0'
+                fiscalyear_ids = [f.id for f in fiscalyears] or [0]
             else:
-                fiscalyear_ids = [int(Transaction().context.get('fiscalyear'))]
-                fiscalyear_clause = '%s' % int(
-                        Transaction().context.get('fiscalyear'))
+                fiscalyear_ids = [Transaction().context.get('fiscalyear')]
 
             if Transaction().context.get('posted'):
-                return (obj + '.state != \'draft\' '
-                    'AND ' + obj + '.move IN ('
-                        'SELECT id FROM account_move '
-                        'WHERE period IN ('
-                            'SELECT id FROM account_period '
-                            'WHERE fiscalyear IN (' + fiscalyear_clause + ')'
-                            ') '
-                            'AND state = \'posted\' '
-                        ')', fiscalyear_ids)
+                return ((table.state != 'draft')
+                    & table.move.in_(
+                        move.select(move.id,
+                            where=move.period.in_(
+                                period.select(period.id,
+                                    where=period.fiscalyear.in_(
+                                        fiscalyear_ids)))
+                            & move.state == 'posted')),
+                    fiscalyear_ids)
             else:
-                return (obj + '.state != \'draft\' '
-                    'AND ' + obj + '.move IN ('
-                        'SELECT id FROM account_move '
-                        'WHERE period IN ('
-                            'SELECT id FROM account_period '
-                            'WHERE fiscalyear IN (' + fiscalyear_clause + ')'
-                            ')'
-                        ')', fiscalyear_ids)
+                return ((table.state != 'draft')
+                    & table.move.in_(
+                        move.select(move.id,
+                            where=move.period.in_(
+                                period.select(period.id,
+                                    where=period.fiscalyear.in_(
+                                        fiscalyear_ids))))),
+                    fiscalyear_ids)
 
     @classmethod
     def on_write(cls, lines):
@@ -1277,7 +1239,7 @@ class Line(ModelSQL, ModelView):
     def write(cls, lines, vals):
         Move = Pool().get('account.move')
 
-        if len(vals) > 1 or 'reconciliation' not in vals:
+        if any(k not in cls._check_modify_exclude for k in vals):
             cls.check_modify(lines)
         moves = [x.move for x in lines]
         super(Line, cls).write(lines, vals)
@@ -1299,15 +1261,6 @@ class Line(ModelSQL, ModelView):
                 if not journal_id:
                     cls.raise_user_error('no_journal')
                 journal = Journal(journal_id)
-                if journal.centralised:
-                    moves = Move.search([
-                            ('period', '=', vals.get('period')
-                                or Transaction().context.get('period')),
-                            ('journal', '=', journal_id),
-                            ('state', '!=', 'posted'),
-                            ], limit=1)
-                    if moves:
-                        vals['move'] = moves[0].id
                 if not vals.get('move'):
                     vals['move'] = Move.create([{
                                 'period': (vals.get('period')
@@ -1448,12 +1401,6 @@ class Line(ModelSQL, ModelView):
                     }])[0]
 
 
-class Move2:
-    __name__ = 'account.move'
-    centralised_line = fields.Many2One('account.move.line', 'Centralised Line',
-            readonly=True)
-
-
 class OpenJournalAsk(ModelView):
     'Open Journal Ask'
     __name__ = 'account.move.open_journal.ask'
@@ -1583,6 +1530,9 @@ class ReconcileLinesWriteOff(ModelView):
     date = fields.Date('Date', required=True)
     account = fields.Many2One('account.account', 'Account', required=True,
             domain=[('kind', '!=', 'view')])
+    amount = fields.Numeric('Amount', digits=(16, Eval('currency_digits', 2)),
+        readonly=True, depends=['currency_digits'])
+    currency_digits = fields.Integer('Currency Digits', readonly=True)
 
     @staticmethod
     def default_date():
@@ -1601,7 +1551,8 @@ class ReconcileLines(Wizard):
             ])
     reconcile = StateTransition()
 
-    def transition_start(self):
+    def get_writeoff(self):
+        "Return writeoff amount and company"
         Line = Pool().get('account.move.line')
 
         company = None
@@ -1610,12 +1561,23 @@ class ReconcileLines(Wizard):
             amount += line.debit - line.credit
             if not company:
                 company = line.account.company
+        return amount, company
+
+    def transition_start(self):
+        amount, company = self.get_writeoff()
         if not company:
             return 'end'
         if company.currency.is_zero(amount):
             return 'reconcile'
         return 'writeoff'
 
+    def default_writeoff(self, fields):
+        amount, company = self.get_writeoff()
+        return {
+            'amount': amount,
+            'currency_digits': company.currency.digits,
+            }
+
     def transition_reconcile(self):
         Line = Pool().get('account.move.line')
 
@@ -1668,10 +1630,18 @@ class OpenReconcileLines(Wizard):
     open_ = StateAction('account.act_move_line_form')
 
     def do_open_(self, action):
-        action['pyson_domain'] = PYSONEncoder().encode([
+        encoder = PYSONEncoder()
+        action['pyson_domain'] = encoder.encode([
             ('account', '=', self.start.account.id),
             ('reconciliation', '=', None),
             ])
+        if self.start.account.kind == 'receivable':
+            order = [('credit', 'DESC'), ('id', 'DESC')]
+        elif self.start.account.kind == 'payable':
+            order = [('debit', 'DESC'), ('id', 'DESC')]
+        else:
+            order = None
+        action['pyson_order'] = encoder.encode(order)
         return action, {}
 
 
